#!/usr/bin/env python3
"""
Tacticalware: OSINT TOOL FOR INSTAGRAM
Implements all commands (including placeholders for emails, phone, etc.)
using instagram_private_api.
"""

import argparse
import sys
import os
import json
import re
from rich.console import Console
from pyfiglet import Figlet
from yaml import safe_load

# instagram_private_api (pip install git+https://github.com/ping/instagram_private_api.git)
from instagram_private_api import (
    Client, ClientError, ClientLoginRequiredError,
    ClientChallengeRequiredError
)

pc = Console()

def log_to_file(message):
    """Log message to a file for analysis."""
    with open("tacticalware_debug.log", "a", encoding="utf-8") as log_file:
        log_file.write(f"{message}\n")

def display_commands(commands):
    """Display available commands."""
    pc.print("Available Commands:", style="bright_blue")
    for command, description in commands.items():
        pc.print(f"  [cyan]{command}[/cyan]: {description}")
    print()

def clear_screen():
    """Clear the terminal screen."""
    os.system('cls' if os.name == 'nt' else 'clear')

def banner():
    """
    Display a simpler banner without the isometric ASCII art.
    """
    clear_screen()
    # A minimal text banner instead of the large isometric ASCII
    pc.print("[bold red]Tacticalware[/bold red]", style="bold red")
    pc.print("OSINT TOOL FOR INSTAGRAM", style="cyan1")
    pc.print("Tacticalware.com", style="bold red")

def graceful_exit():
    """Exit the program gracefully."""
    pc.print("Thank you for using Tacticalware!", style="yellow")
    sys.exit(0)

def handle_error(e):
    """Handle and display errors."""
    error_message = f"[red]Error:[/red] {e}"
    pc.print(error_message, style="orange1")
    log_to_file(f"Error: {e}")

def load_credentials(file_path):
    """Load credentials from a YAML file."""
    try:
        with open(file_path, 'r', encoding='utf-8') as file:
            return safe_load(file)
    except FileNotFoundError:
        error_message = f"Credentials file '{file_path}' not found."
        pc.print(f"[red]Error:[/red] {error_message}", style="red")
        log_to_file(f"Error: {error_message}")
        sys.exit(1)
    except Exception as e:
        handle_error(e)
        sys.exit(1)

def handle_challenge(api, challenge_exc):
    """
    Handle 'challenge_required' flow from Instagram.
    Prompts user for verification method and code if needed.
    """
    challenge_url = None
    if challenge_exc.error_response:
        challenge_info = challenge_exc.error_response.get('challenge', {})
        challenge_url = challenge_info.get('api_path')

    if not challenge_url:
        pc.print("[red]Error:[/red] Could not find challenge URL.", style="red")
        return False

    pc.print(f"[yellow]Challenge required. Trying to resolve at: {challenge_url}[/yellow]")

    try:
        # Step 1: See available methods
        challenge = api.challenge_resolve(challenge_url)
        pc.print(f"Challenge details: {challenge}", style="yellow")

        step_data = challenge.get('step_data', {})
        step_name = challenge.get('step_name', '')

        # If multiple methods, prompt user
        if step_name == 'select_verify_method':
            choice = None
            if step_data.get('choice'):
                pc.print("Challenge: choose verification method:", style="cyan")
                for key in step_data:
                    pc.print(f" - {key}: {step_data[key]}")
                pc.print("Which method do you want to use? (phone_number or email)", style="yellow")
                choice = input("Method choice: ").strip()
            if choice not in ('phone_number', 'email'):
                pc.print("Defaulting to 'phone_number'", style="yellow")
                choice = 'phone_number'

            api.challenge_select_verify_method(challenge_url, choice=choice)
            pc.print(f"Verification method set to '{choice}'. Check your device for the code.", style="green")

        elif step_name in ('verify_code', 'verify_email', 'verify_phone'):
            pc.print("Verify code step. You should receive a code via phone/email.", style="cyan")

        # Step 2: Prompt for code
        code = input("Enter the verification code sent to you: ").strip()
        if not code:
            pc.print("[red]No code entered.[/red]", style="red")
            return False

        # Step 3: Send code
        final_challenge = api.challenge_send_security_code(challenge_url, code)
        pc.print(f"Challenge final response: {final_challenge}", style="yellow")

        if final_challenge.get('status') == 'ok':
            pc.print("[green]Challenge passed! You are now logged in.[/green]")
            return True
        else:
            pc.print("[red]Challenge verification failed.[/red]", style="red")
            return False

    except Exception as e:
        handle_error(f"Challenge error: {e}")
        return False

def instagram_login(creds):
    """
    Login to Instagram with the Private API (handles challenge if needed).
    Returns an authenticated API client.
    """
    pc.print("Attempting to Login :)", style="cyan")
    from instagram_private_api import Client, ClientError, ClientLoginRequiredError, ClientChallengeRequiredError
    try:
        api = Client(creds['username'], creds['password'])
        pc.print("Login successful!", style="green")
        return api
    except ClientChallengeRequiredError as c:
        pc.print("[yellow]Instagram requires a challenge. Attempting to handle...[/yellow]")
        # Re-instantiate partially so we can handle the challenge
        api = Client(creds['username'], creds['password'], auto_patch=True, authenticate=False)
        if handle_challenge(api, c):
            pc.print("Challenge completed. You should now be logged in.", style="green")
            return api
        else:
            pc.print("[red]Challenge handling failed.[/red]", style="red")
            sys.exit(1)
    except (ClientError, ClientLoginRequiredError) as e:
        pc.print(f"Login failed: {e}", style="red")
        log_to_file(f"Login failed: {e}")
        sys.exit(1)
    except Exception as e:
        handle_error(f"Unexpected error during login: {e}")
        sys.exit(1)

def get_user_id(api, username):
    """
    Retrieve the user ID (pk) for a given username.
    """
    try:
        info = api.username_info(username)
        return info['user']['pk']
    except Exception as e:
        handle_error(f"Failed to get user ID for {username}: {e}")
        return None


# --------------------------------------------------------------------
# COMMAND IMPLEMENTATIONS
# --------------------------------------------------------------------

def download_stories(api, target):
    """Download Instagram stories for the target user."""
    pc.print(f"Fetching stories for {target}...", style="cyan")
    user_id = get_user_id(api, target)
    if not user_id:
        return

    try:
        reel = api.user_reel_media(user_id)
        if reel and reel.get('items'):
            os.makedirs(f"stories_{target}", exist_ok=True)
            for item in reel['items']:
                media_url = None
                if 'video_versions' in item:
                    media_url = item['video_versions'][0]['url']
                elif 'image_versions2' in item:
                    candidates = item['image_versions2'].get('candidates', [])
                    if candidates:
                        media_url = candidates[0]['url']
                if media_url:
                    pc.print(f"Downloading story from {media_url}", style="yellow")
                    media_data = api.session.get(media_url).content
                    filename = (
                        f"stories_{target}/{item['id']}.mp4"
                        if 'video_versions' in item else
                        f"stories_{target}/{item['id']}.jpg"
                    )
                    with open(filename, "wb") as f:
                        f.write(media_data)
                    pc.print(f"Saved story to {filename}", style="green")
            pc.print("All available stories downloaded.", style="green")
        else:
            pc.print("No stories found for the target.", style="yellow")
    except Exception as e:
        handle_error(f"Failed to fetch stories for {target}: {e}")


def get_followers(api, target):
    """Fetch a list of the target's followers."""
    pc.print(f"Fetching followers for {target}...", style="cyan")
    user_id = get_user_id(api, target)
    if not user_id:
        return []

    rank_token = api.generate_uuid()
    followers_list = []

    try:
        results = api.user_followers(user_id, rank_token=rank_token)
        followers = results.get('users', [])
        followers_list.extend([u.get('username') for u in followers])

        next_max_id = results.get('next_max_id')
        while next_max_id:
            results = api.user_followers(user_id, rank_token=rank_token, max_id=next_max_id)
            followers = results.get('users', [])
            followers_list.extend([u.get('username') for u in followers])
            next_max_id = results.get('next_max_id')

        pc.print("Followers:", style="bright_blue")
        for f in followers_list:
            pc.print(f" - {f}")
        return followers_list
    except Exception as e:
        handle_error(f"Failed to fetch followers: {e}")
    return []


def get_following(api, target):
    """Fetch a list of accounts that the target is following."""
    pc.print(f"Fetching following list for {target}...", style="cyan")
    user_id = get_user_id(api, target)
    if not user_id:
        return []

    rank_token = api.generate_uuid()
    following_list = []

    try:
        results = api.user_following(user_id, rank_token=rank_token)
        following = results.get('users', [])
        following_list.extend([u.get('username') for u in following])

        next_max_id = results.get('next_max_id')
        while next_max_id:
            results = api.user_following(user_id, rank_token=rank_token, max_id=next_max_id)
            following = results.get('users', [])
            following_list.extend([u.get('username') for u in following])
            next_max_id = results.get('next_max_id')

        pc.print("Following:", style="bright_blue")
        for f in following_list:
            pc.print(f" - {f}")
        return following_list
    except Exception as e:
        handle_error(f"Failed to fetch following: {e}")
    return []


def get_followers_emails(api, target):
    """
    Retrieve email addresses of the target's followers by parsing their bios.
    This is not official data—just scanning text for email patterns.
    """
    pc.print(f"Fetching followers' emails for {target}...", style="cyan")
    followers_list = get_followers(api, target)
    if not followers_list:
        return {}

    emails_dict = {}
    email_pattern = re.compile(r"[a-zA-Z0-9_.+\-]+@[a-zA-Z0-9\-]+\.[a-zA-Z0-9\-.]+")

    for follower_username in followers_list:
        try:
            user_info = api.username_info(follower_username)
            bio = user_info['user'].get('biography', '')
            found_emails = email_pattern.findall(bio)
            emails_dict[follower_username] = list(set(found_emails)) if found_emails else []
        except Exception as e:
            handle_error(f"Failed to get bio/email for {follower_username}: {e}")

    for user, email_list in emails_dict.items():
        pc.print(f"\nFollower: {user}", style="bright_blue")
        if email_list:
            pc.print(f"  Emails found: {email_list}", style="green")
        else:
            pc.print("  No emails found in bio.", style="yellow")

    return emails_dict


def get_following_emails(api, target):
    """
    Retrieve email addresses of users the target follows by scanning their bios.
    """
    pc.print(f"Fetching following emails for {target}...", style="cyan")
    following_list = get_following(api, target)
    if not following_list:
        return {}

    emails_dict = {}
    email_pattern = re.compile(r"[a-zA-Z0-9_.+\-]+@[a-zA-Z0-9\-]+\.[a-zA-Z0-9\-.]+")

    for username in following_list:
        try:
            user_info = api.username_info(username)
            bio = user_info['user'].get('biography', '')
            found_emails = email_pattern.findall(bio)
            emails_dict[username] = list(set(found_emails)) if found_emails else []
        except Exception as e:
            handle_error(f"Failed to get bio/email for {username}: {e}")

    for user, email_list in emails_dict.items():
        pc.print(f"\nFollowing: {user}", style="bright_blue")
        if email_list:
            pc.print(f"  Emails found: {email_list}", style="green")
        else:
            pc.print("  No emails found in bio.", style="yellow")

    return emails_dict


def get_followers_phone(api, target):
    """
    Retrieve phone numbers of the target's followers by scanning their bios.
    """
    pc.print(f"Fetching followers' phone numbers for {target}...", style="cyan")
    followers_list = get_followers(api, target)
    if not followers_list:
        return {}

    phones_dict = {}
    phone_pattern = re.compile(r"(\+?\d[\d .-]{7,}\d)")

    for follower_username in followers_list:
        try:
            user_info = api.username_info(follower_username)
            bio = user_info['user'].get('biography', '')
            found_phones = phone_pattern.findall(bio)
            phones_dict[follower_username] = list(set(found_phones)) if found_phones else []
        except Exception as e:
            handle_error(f"Failed to get phone for {follower_username}: {e}")

    for user, phone_list in phones_dict.items():
        pc.print(f"\nFollower: {user}", style="bright_blue")
        if phone_list:
            pc.print(f"  Possible phone numbers: {phone_list}", style="green")
        else:
            pc.print("  No phone numbers found in bio.", style="yellow")

    return phones_dict


def get_following_phone(api, target):
    """
    Retrieve phone numbers of the users the target follows by scanning their bios.
    """
    pc.print(f"Fetching following phone numbers for {target}...", style="cyan")
    following_list = get_following(api, target)
    if not following_list:
        return {}

    phones_dict = {}
    phone_pattern = re.compile(r"(\+?\d[\d .-]{7,}\d)")

    for username in following_list:
        try:
            user_info = api.username_info(username)
            bio = user_info['user'].get('biography', '')
            found_phones = phone_pattern.findall(bio)
            phones_dict[username] = list(set(found_phones)) if found_phones else []
        except Exception as e:
            handle_error(f"Failed to get phone for {username}: {e}")

    for user, phone_list in phones_dict.items():
        pc.print(f"\nFollowing: {user}", style="bright_blue")
        if phone_list:
            pc.print(f"  Possible phone numbers: {phone_list}", style="green")
        else:
            pc.print("  No phone numbers found in bio.", style="yellow")

    return phones_dict


def get_likes(api, target):
    """Fetch the likes (usernames) for the target's recent posts."""
    pc.print(f"Fetching likes for recent posts from {target}...", style="cyan")
    user_id = get_user_id(api, target)
    if not user_id:
        return {}

    all_likes = {}
    try:
        results = api.user_feed(user_id)
        items = results.get('items', [])
        for item in items:
            media_id = item.get('id')
            if not media_id:
                continue
            likes_info = api.media_likers(media_id)
            users = likes_info.get('users', [])
            usernames = [u.get('username') for u in users]
            all_likes[media_id] = usernames

        for media, liker_list in all_likes.items():
            pc.print(f"\nLikes for post ID {media}:", style="bright_blue")
            for user in liker_list:
                pc.print(f"  - {user}")
        return all_likes

    except Exception as e:
        handle_error(f"Failed to fetch likes: {e}")
    return {}


def get_mediatype(api, target):
    """
    Identify if the user's recent posts are photos, videos, or carousels.
    We'll list each post ID and its type (1=photo, 2=video, 8=carousel).
    """
    pc.print(f"Fetching media types for {target}...", style="cyan")
    user_id = get_user_id(api, target)
    if not user_id:
        return {}

    media_types = {}
    try:
        results = api.user_feed(user_id)
        items = results.get('items', [])
        for item in items:
            media_id = item.get('id')
            m_type = item.get('media_type')
            if m_type == 1:
                type_str = "Photo"
            elif m_type == 2:
                type_str = "Video"
            elif m_type == 8:
                type_str = "Carousel/Album"
            else:
                type_str = f"Unknown({m_type})"
            media_types[media_id] = type_str

        for mid, tstr in media_types.items():
            pc.print(f"Post ID: {mid} -> {tstr}", style="bright_blue")

        return media_types
    except Exception as e:
        handle_error(f"Failed to fetch media types: {e}")
    return {}


def get_captions(api, target):
    """Fetch captions of the target's recent posts."""
    pc.print(f"Fetching captions for {target}...", style="cyan")
    user_id = get_user_id(api, target)
    if not user_id:
        return {}

    captions_dict = {}
    try:
        results = api.user_feed(user_id)
        items = results.get('items', [])
        for item in items:
            media_id = item.get('id')
            caption_text = (item.get('caption') or {}).get('text', "")
            captions_dict[media_id] = caption_text

        for mid, cap in captions_dict.items():
            pc.print(f"\nPost ID: {mid}", style="bright_blue")
            pc.print(f"Caption: {cap}", style="green")

        return captions_dict
    except Exception as e:
        handle_error(f"Failed to fetch captions: {e}")
    return {}


def get_comments(api, target):
    """
    Retrieve the textual comments for each of the target's recent posts.
    Different from 'commenter' which only lists commenters' usernames.
    """
    pc.print(f"Fetching comments for {target}...", style="cyan")
    user_id = get_user_id(api, target)
    if not user_id:
        return {}

    all_comments = {}
    try:
        results = api.user_feed(user_id)
        items = results.get('items', [])
        for item in items:
            media_id = item.get('id')
            if not media_id:
                continue
            try:
                comments_info = api.media_n_comments(media_id, n=100)
            except Exception as e:
                handle_error(f"Failed to fetch comments for post {media_id}: {e}")
                continue
            comments_list = comments_info.get('comments', [])
            texts = []
            for c in comments_list:
                user_ = c.get('user', {}).get('username', '')
                text_ = c.get('text', '')
                texts.append((user_, text_))
            all_comments[media_id] = texts

        for mid, c_list in all_comments.items():
            pc.print(f"\nComments for post {mid}:", style="bright_blue")
            for user_, text_ in c_list:
                pc.print(f"  {user_}: {text_}")

        return all_comments
    except Exception as e:
        handle_error(f"Failed to fetch comments: {e}")
    return {}


def get_commenters(api, target):
    """Fetch the commenters (usernames) for the target's recent posts."""
    pc.print(f"Fetching commenters for recent posts from {target}...", style="cyan")
    user_id = get_user_id(api, target)
    if not user_id:
        return {}

    all_commenters = {}
    try:
        results = api.user_feed(user_id)
        items = results.get('items', [])
        for item in items:
            media_id = item.get('id')
            if not media_id:
                continue
            try:
                comments_info = api.media_n_comments(media_id, n=100)
            except Exception as e:
                handle_error(f"Failed to fetch comments for post {media_id}: {e}")
                continue
            comments_list = comments_info.get('comments', [])
            commenter_usernames = [c.get('user', {}).get('username') for c in comments_list]
            all_commenters[media_id] = commenter_usernames

        for media, comm_list in all_commenters.items():
            pc.print(f"\nCommenters for post ID {media}:", style="bright_blue")
            for commenter in comm_list:
                pc.print(f"  - {commenter}")

        return all_commenters
    except Exception as e:
        handle_error(f"Failed to fetch commenters: {e}")
    return {}


def download_photos(api, target):
    """
    Download the user's recent photo posts (media_type=1 or photos in a carousel).
    Excludes videos for simplicity.
    """
    pc.print(f"Downloading photos for {target}...", style="cyan")
    user_id = get_user_id(api, target)
    if not user_id:
        return

    try:
        results = api.user_feed(user_id)
        items = results.get('items', [])
        os.makedirs(f"photos_{target}", exist_ok=True)

        for item in items:
            m_type = item.get('media_type')
            media_id = item.get('id')
            if m_type == 1:
                # single photo
                versions = item.get('image_versions2', {}).get('candidates', [])
                if versions:
                    url = versions[0].get('url')
                    if url:
                        pc.print(f"Downloading photo from {url}", style="yellow")
                        data = api.session.get(url).content
                        filename = f"photos_{target}/{media_id}.jpg"
                        with open(filename, 'wb') as f:
                            f.write(data)
                        pc.print(f"Saved photo to {filename}", style="green")
            elif m_type == 8:
                # Carousel
                carousel = item.get('carousel_media', [])
                for idx, media_item in enumerate(carousel, start=1):
                    sub_type = media_item.get('media_type')
                    if sub_type == 1:  # photo
                        versions = media_item.get('image_versions2', {}).get('candidates', [])
                        if versions:
                            url = versions[0].get('url')
                            if url:
                                pc.print(f"Downloading carousel photo from {url}", style="yellow")
                                data = api.session.get(url).content
                                filename = f"photos_{target}/{media_id}_{idx}.jpg"
                                with open(filename, 'wb') as f:
                                    f.write(data)
                                pc.print(f"Saved photo to {filename}", style="green")
            # skip video posts
        pc.print("Photo download completed.", style="green")
    except Exception as e:
        handle_error(f"Failed to download photos: {e}")


def download_profile_pic(api, target):
    """Download the target's profile picture."""
    pc.print(f"Downloading profile picture for {target}...", style="cyan")
    user_id = get_user_id(api, target)
    if not user_id:
        return

    try:
        info = api.user_info(user_id)
        user_data = info.get('user', {})
        pp_url = user_data.get('hd_profile_pic_url_info', {}).get('url')
        if not pp_url:
            pp_url = user_data.get('profile_pic_url')

        if pp_url:
            pc.print(f"Downloading profile pic from {pp_url}", style="yellow")
            data = api.session.get(pp_url).content
            os.makedirs(f"profile_{target}", exist_ok=True)
            filename = f"profile_{target}/{target}_profile_pic.jpg"
            with open(filename, 'wb') as f:
                f.write(data)
            pc.print(f"Saved profile picture to {filename}", style="green")
        else:
            pc.print("No profile picture URL found.", style="yellow")
    except Exception as e:
        handle_error(f"Failed to download profile picture: {e}")


def get_tags(api, target):
    """
    Fetch hashtags used by the target's recent posts.
    """
    pc.print(f"Fetching tags (hashtags) for {target}...", style="cyan")
    user_id = get_user_id(api, target)
    if not user_id:
        return []

    tags_found = set()
    try:
        results = api.user_feed(user_id)
        items = results.get('items', [])
        for item in items:
            caption = (item.get('caption') or {}).get('text', "")
            words = caption.split()
            for w in words:
                if w.startswith('#'):
                    tag = w.strip('#.,!?:;')
                    tags_found.add(tag)

        next_max_id = results.get('next_max_id')
        while next_max_id:
            results = api.user_feed(user_id, max_id=next_max_id)
            items = results.get('items', [])
            for item in items:
                caption = (item.get('caption') or {}).get('text', "")
                words = caption.split()
                for w in words:
                    if w.startswith('#'):
                        tag = w.strip('#.,!?:;')
                        tags_found.add(tag)
            next_max_id = results.get('next_max_id')

        tags_list = sorted(list(tags_found))
        pc.print("Hashtags found:", style="bright_blue")
        for t in tags_list:
            pc.print(f" - #{t}")
        return tags_list
    except Exception as e:
        handle_error(f"Failed to fetch tags: {e}")
    return []


def get_tagged(api, target):
    """
    Get a list of media posts in which the target is tagged (usertags).
    """
    pc.print(f"Fetching media where {target} is tagged...", style="cyan")
    user_id = get_user_id(api, target)
    if not user_id:
        return []

    tagged_posts = []
    try:
        results = api.usertag_feed(user_id)
        items = results.get('items', [])
        for item in items:
            media_id = item.get('id')
            caption_text = (item.get('caption') or {}).get('text', "")
            tagged_posts.append({"id": media_id, "caption": caption_text})

        next_max_id = results.get('next_max_id')
        while next_max_id:
            results = api.usertag_feed(user_id, max_id=next_max_id)
            items = results.get('items', [])
            for item in items:
                media_id = item.get('id')
                caption_text = (item.get('caption') or {}).get('text', "")
                tagged_posts.append({"id": media_id, "caption": caption_text})
            next_max_id = results.get('next_max_id')

        for tp in tagged_posts:
            pc.print(f"Post ID: {tp['id']} - Caption: {tp['caption']}", style="bright_blue")
        return tagged_posts

    except Exception as e:
        handle_error(f"Failed to fetch tagged media: {e}")
    return []


def get_timeline(api, target):
    """
    Retrieve a basic timeline (recent feed) for the target:
    post info (IDs, short captions, timestamps) for the user's feed.
    """
    pc.print(f"Fetching timeline for {target}...", style="cyan")
    user_id = get_user_id(api, target)
    if not user_id:
        return []

    timeline_posts = []
    try:
        results = api.user_feed(user_id)
        items = results.get('items', [])
        for item in items:
            media_id = item.get('id')
            caption_text = (item.get('caption') or {}).get('text', "")
            taken_at = item.get('taken_at')
            post_info = {
                "id": media_id,
                "caption": caption_text,
                "timestamp": taken_at,
            }
            timeline_posts.append(post_info)

        for post in timeline_posts:
            pc.print(f"\nPost ID: {post['id']}", style="bright_blue")
            pc.print(f"Caption: {post['caption']}", style="green")
            pc.print(f"Timestamp: {post['timestamp']}", style="green")

        return timeline_posts
    except Exception as e:
        handle_error(f"Failed to fetch timeline: {e}")
    return []


# --------------------------------------------------------------------
# MAIN PROCESSING LOOP
# --------------------------------------------------------------------
def process_instagram(target, creds, save_json, save_file):
    """Process Instagram commands using the private API (with all features)."""
    api = instagram_login(creds)
    if not api:
        return

    commands = {
        'followers': "Get target's followers",
        'following': "Get accounts target follows",
        'followers emails': "Retrieve email addresses of the target's followers",
        'following emails': "Retrieve email addresses of users the target follows",
        'followers phone': "Retrieve phone numbers of the target's followers",
        'following phone': "Retrieve phone numbers of users the target follows",
        'likes': "Get likes on target's posts",
        'mediatype': "Identify if posts are photos or videos",
        'captions': "Fetch captions of the target's posts",
        'comments': "Retrieve comments on the target's posts",
        'commenter': "Get a list of users who commented on the target's posts",
        'stories': "Download the target's stories",
        'photos': "Download the target's photos",
        'profile_pic': "Download the target's profile picture",
        'tags': "Fetch hashtags used by the target",
        'tagged': "Get a list of users tagged by the target",
        'timeline': "Retrieve detailed account timeline",
        'reset target': "Reset the target",
    }

    display_commands(commands)
    while True:
        pc.print("~/Tacticalware Instagram >$ ", style='purple', end='')
        user_input = input().strip()
        if user_input == 'exit':
            graceful_exit()
        elif user_input == 'clear':
            clear_screen()
        elif user_input == 'help':
            display_commands(commands)
        elif user_input == 'reset target':
            pc.print("Resetting target. Please restart the program with a new target.", style="yellow")
            log_to_file("Target reset by user.")
            break
        elif user_input in commands:
            pc.print(f"Executing: {user_input}", style="green")
            log_to_file(f"Command executed: {user_input}")

            # 1. followers
            if user_input == 'followers':
                followers_list = get_followers(api, target)
                if save_json and followers_list:
                    with open(f"{target}_followers.json", "w", encoding="utf-8") as f:
                        json.dump(followers_list, f, ensure_ascii=False, indent=4)
                if save_file and followers_list:
                    with open(f"{target}_followers.txt", "w", encoding="utf-8") as f:
                        for follower in followers_list:
                            f.write(f"{follower}\n")

            # 2. following
            elif user_input == 'following':
                following_list = get_following(api, target)
                if save_json and following_list:
                    with open(f"{target}_following.json", "w", encoding="utf-8") as f:
                        json.dump(following_list, f, ensure_ascii=False, indent=4)
                if save_file and following_list:
                    with open(f"{target}_following.txt", "w", encoding="utf-8") as f:
                        for user in following_list:
                            f.write(f"{user}\n")

            # 3. followers emails
            elif user_input == 'followers emails':
                emails_dict = get_followers_emails(api, target)
                if save_json and emails_dict:
                    with open(f"{target}_followers_emails.json", "w", encoding="utf-8") as f:
                        json.dump(emails_dict, f, ensure_ascii=False, indent=4)
                if save_file and emails_dict:
                    with open(f"{target}_followers_emails.txt", "w", encoding="utf-8") as f:
                        for user, mail_list in emails_dict.items():
                            f.write(f"{user}: {mail_list}\n")

            # 4. following emails
            elif user_input == 'following emails':
                emails_dict = get_following_emails(api, target)
                if save_json and emails_dict:
                    with open(f"{target}_following_emails.json", "w", encoding="utf-8") as f:
                        json.dump(emails_dict, f, ensure_ascii=False, indent=4)
                if save_file and emails_dict:
                    with open(f"{target}_following_emails.txt", "w", encoding="utf-8") as f:
                        for user, mail_list in emails_dict.items():
                            f.write(f"{user}: {mail_list}\n")

            # 5. followers phone
            elif user_input == 'followers phone':
                phone_dict = get_followers_phone(api, target)
                if save_json and phone_dict:
                    with open(f"{target}_followers_phone.json", "w", encoding="utf-8") as f:
                        json.dump(phone_dict, f, ensure_ascii=False, indent=4)
                if save_file and phone_dict:
                    with open(f"{target}_followers_phone.txt", "w", encoding="utf-8") as f:
                        for user, p_list in phone_dict.items():
                            f.write(f"{user}: {p_list}\n")

            # 6. following phone
            elif user_input == 'following phone':
                phone_dict = get_following_phone(api, target)
                if save_json and phone_dict:
                    with open(f"{target}_following_phone.json", "w", encoding="utf-8") as f:
                        json.dump(phone_dict, f, ensure_ascii=False, indent=4)
                if save_file and phone_dict:
                    with open(f"{target}_following_phone.txt", "w", encoding="utf-8") as f:
                        for user, p_list in phone_dict.items():
                            f.write(f"{user}: {p_list}\n")

            # 7. likes
            elif user_input == 'likes':
                likes_dict = get_likes(api, target)
                if save_json and likes_dict:
                    with open(f"{target}_likes.json", "w", encoding="utf-8") as f:
                        json.dump(likes_dict, f, ensure_ascii=False, indent=4)
                if save_file and likes_dict:
                    with open(f"{target}_likes.txt", "w", encoding="utf-8") as f:
                        for media_id, liker_list in likes_dict.items():
                            f.write(f"Post ID {media_id}:\n")
                            for name in liker_list:
                                f.write(f"  - {name}\n")
                            f.write("\n")

            # 8. mediatype
            elif user_input == 'mediatype':
                media_types = get_mediatype(api, target)
                if save_json and media_types:
                    with open(f"{target}_mediatype.json", "w", encoding="utf-8") as f:
                        json.dump(media_types, f, ensure_ascii=False, indent=4)
                if save_file and media_types:
                    with open(f"{target}_mediatype.txt", "w", encoding="utf-8") as f:
                        for mid, tstr in media_types.items():
                            f.write(f"{mid}: {tstr}\n")

            # 9. captions
            elif user_input == 'captions':
                captions_dict = get_captions(api, target)
                if save_json and captions_dict:
                    with open(f"{target}_captions.json", "w", encoding="utf-8") as f:
                        json.dump(captions_dict, f, ensure_ascii=False, indent=4)
                if save_file and captions_dict:
                    with open(f"{target}_captions.txt", "w", encoding="utf-8") as f:
                        for mid, cap in captions_dict.items():
                            f.write(f"Post ID {mid}:\n{cap}\n\n")

            # 10. comments
            elif user_input == 'comments':
                comments_dict = get_comments(api, target)
                if save_json and comments_dict:
                    with open(f"{target}_comments.json", "w", encoding="utf-8") as f:
                        json.dump(comments_dict, f, ensure_ascii=False, indent=4)
                if save_file and comments_dict:
                    with open(f"{target}_comments.txt", "w", encoding="utf-8") as f:
                        for mid, c_list in comments_dict.items():
                            f.write(f"Comments for post {mid}:\n")
                            for user_, text_ in c_list:
                                f.write(f"  {user_}: {text_}\n")
                            f.write("\n")

            # 11. commenter
            elif user_input == 'commenter':
                commenters_dict = get_commenters(api, target)
                if save_json and commenters_dict:
                    with open(f"{target}_commenters.json", "w", encoding="utf-8") as f:
                        json.dump(commenters_dict, f, ensure_ascii=False, indent=4)
                if save_file and commenters_dict:
                    with open(f"{target}_commenters.txt", "w", encoding="utf-8") as f:
                        for mid, comm_list in commenters_dict.items():
                            f.write(f"Post ID {mid}:\n")
                            for user_ in comm_list:
                                f.write(f"  - {user_}\n")
                            f.write("\n")

            # 12. stories
            elif user_input == 'stories':
                download_stories(api, target)

            # 13. photos
            elif user_input == 'photos':
                download_photos(api, target)

            # 14. profile_pic
            elif user_input == 'profile_pic':
                download_profile_pic(api, target)

            # 15. tags
            elif user_input == 'tags':
                tags_list = get_tags(api, target)
                if save_json and tags_list:
                    with open(f"{target}_tags.json", "w", encoding="utf-8") as f:
                        json.dump(tags_list, f, ensure_ascii=False, indent=4)
                if save_file and tags_list:
                    with open(f"{target}_tags.txt", "w", encoding="utf-8") as f:
                        for t in tags_list:
                            f.write(f"#{t}\n")

            # 16. tagged
            elif user_input == 'tagged':
                tagged_list = get_tagged(api, target)
                if save_json and tagged_list:
                    with open(f"{target}_tagged.json", "w", encoding="utf-8") as f:
                        json.dump(tagged_list, f, ensure_ascii=False, indent=4)
                if save_file and tagged_list:
                    with open(f"{target}_tagged.txt", "w", encoding="utf-8") as f:
                        for item in tagged_list:
                            f.write(f"ID: {item['id']} - {item['caption']}\n")

            # 17. timeline
            elif user_input == 'timeline':
                timeline_list = get_timeline(api, target)
                if save_json and timeline_list:
                    with open(f"{target}_timeline.json", "w", encoding="utf-8") as f:
                        json.dump(timeline_list, f, ensure_ascii=False, indent=4)
                if save_file and timeline_list:
                    with open(f"{target}_timeline.txt", "w", encoding="utf-8") as f:
                        for post in timeline_list:
                            f.write(f"Post ID: {post['id']}\n")
                            f.write(f"Caption: {post['caption']}\n")
                            f.write(f"Timestamp: {post['timestamp']}\n\n")

        else:
            if user_input.strip():
                invalid_command_message = (
                    f"Invalid command '{user_input}'. Type 'help' for a list of commands."
                )
                pc.print(invalid_command_message, style="red")
                log_to_file(invalid_command_message)


def main():
    """Main function to handle user input and arguments."""
    banner()
    parser = argparse.ArgumentParser(description="OSINT Tool for Instagram")
    parser.add_argument('target', type=str, help="Target username (without https://...)")
    parser.add_argument('-j', '--json', help="Save results in a JSON file", action='store_true')
    parser.add_argument('-f', '--file', help="Save results in a text file", action='store_true')
    args = parser.parse_args()

    creds = load_credentials("creds/insta.yml")
    
    pc.print("Choose a platform:", style="cyan")
    pc.print("  1. Instagram", style="green3")
    pc.print("  2. Exit", style="red")
    pc.print("> Select an option: ", style='purple', end='')
    choice = input().strip()
    if choice == '1':
        process_instagram(args.target, creds, args.json, args.file)
    elif choice == '2':
        graceful_exit()
    else:
        invalid_choice_message = "Invalid choice. Exiting..."
        pc.print(invalid_choice_message, style="red")
        log_to_file(invalid_choice_message)
        sys.exit(1)

if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        interruption_message = "\nProgram interrupted. Exiting..."
        pc.print(interruption_message, style="bold red")
        log_to_file(interruption_message)
        sys.exit(0)
    except Exception as e:
        handle_error(e)
        sys.exit(1)
